module Exercise_10 (module Join, module SafeMap) where

{- Library -- nicht veraendern -}
import SafeMap (SafeMap, empty, update, lookup)
import Join (Field, doJoin, doJoinFormat, parseFormat)

{- G10.1 -}

{-
\alpha ist wie in den Folien definiert.

Die ersten zwei Eigenschaften lassen sich direkt beweisen:

    \alpha empty
      = \alpha []                           -- by def of empty
      = {}                                  -- by def of \alpha

    \alpha (insert x xs)
      = \alpha (x : xs)                     -- by def of insert
      = {x} \union \alpha xs                -- by def of \alpha

Fuer die dritte Eigenschaft wird das folgende Lemma gebraucht:

    elem x xs = x \in \alpha xs

Beweis durch strukturelle Induktion auf "xs":

    Induktionsanfang:
      Zu beweisen: elem x [] = x \in \alpha []

      elem x [] = False                     -- by def of elem
      x \in \alpha []
        = x \in {}                          -- by def of \alpha
        = False                             -- by basic property of \in

    Induktionsschritt:
      IH: elem x ys = x \in \alpha ys
      Zu beweisen: elem x (y : ys) = x \in \alpha (y : ys)

      elem x (y : ys)
        = x == y || elem x ys               -- by def of elem
        = x == y || x \in \alpha ys         -- by IH

      x \in \alpha (y : ys)
        = x \in {y} \union \alpha ys        -- by def of \alpha
        = x == y || x \in \alpha ys         -- by basic property of \union

Daraus folgt

    isin x xs = elem x xs = x \in \alpha xs

Fuer die vierte Eigenschaft wird das folgende Lemma gebraucht:

    length (nub xs) = |\alpha xs|

Beweis durch strukturelle Induktion auf "xs":

    Induktionsanfang:
      Zu beweisen: length (nub []) = |\alpha []|

      length (nub [])
        = length []                         -- by def of nub
        = 0                                 -- by def of length

      |\alpha []|
        = |{}|                              -- by def of \alpha
        = 0                                 -- by basic propery of | |

    Induktionsschritt:
      IH: length (nub xs) = |\alpha xs|
      Zu beweisen: length (nub (x : xs)) = |\alpha (x : xs)|

      Fallunterscheidung auf x `elem` xs:

      Fall x `elem` xs:
        length (nub (x : xs))
          = length (nub xs)                 -- by length_nub_dup
          = |\alpha xs|                     -- by IH
          = |\alpha (x : xs)|               -- by alpha_dup

      Fall not (x `elem` xs):
        length (nub (x : xs))
          = length (nub xs) + 1             -- by length_nub_nodup
          = |\alpha xs| + 1                 -- by IH
          = |\alpha (x : xs)|               -- by alpha_nodup

Daraus folgt

    size xs = length (nub xs) = |\alpha xs|

Die Lemmas length_nub_dup, length_nub_nodup, alpha_dup, alpha_nodup sollte man
im Prinzip auch beweisen, aber der Beweis ist schon lang genug ohne. ;)
-}


{- G10.2: siehe SetPairList.hs -}


{- G10.3: siehe Queue.hs und SuperQueue.hs -}

{-
Simulationseigenschaften zu beweisen:

    \alpha empty = []
    isEmpty q = \alpha q == []
    \alpha (enqueue x q) = \alpha q ++ [x]
    ~ isEmpty q ==> \alpha (dequeue q) = (head (\alpha q), tail (\alpha q))
    q == q' = \alpha q == \alpha q'

Definition von \alpha:

    \alpha (Queue front kcab) = front ++ reverse kcab

Beweis fuer "empty":

    \alpha empty
      = \alpha (Queue [] [])                -- by def of empty
      = [] ++ reverse []                    -- by def of \alpha
      = []                                  -- by def of reverse, ++

Beweis fuer "isEmpty":

    Hilfssatz: q == Queue [] [] = \alpha q == []

    isEmpty q
      = q == Queue [] []                    -- by def of isEmpty
      = \alpha q == []                      -- by lemma

Beweis fuer "enqueue":

    Sei q = Queue front kcab                -- by exhaustivity of constructor

    \alpha (enqueue x q)
      = \alpha (enqueue x (Queue front kcab))
                                            -- by def of q
      = \alpha (Queue (front (x : kcab)))   -- by def of enqueue
      = front ++ reverse (x : kcab)         -- by def of \alpha
      = front ++ (reverse kcab ++ [x])      -- by def of reverse
      = (front ++ reverse kcab) ++ [x]      -- by commutativity of ++
      = \alpha (Queue front kcab) ++ [x]    -- by def of \alpha
      = \alpha q ++ [x]                     -- by def of q

Hilfslemma L:

    Zu beweisen: \alpha (Queue [] kcab) = \alpha (Queue (reverse kcab) [])

    \alpha (Queue [] kcab)
      = [] ++ reverse kcab                  -- by def of \alpha
      = reverse kcab                        -- by def of ++

    \alpha (Queue (reverse kcab) [])
      = reverse kcab ++ []                  -- by def of \alpha
      = reverse kcab                        -- by def of ++

Beweis fuer "dequeue":

    Annahme: ~ isEmpty q

    Sei q = Queue front kcab

    Fallunterscheidung auf front und kcab:

    Fall front = [], kcab = []:
      ~ isEmpty (Queue front kcab)
        = ~ isEmpty (Queue [] [])
        = False
      Nichts mehr zu beweisen, weil die Annahme nicht erfuellt ist.

    Fall front /= []:
      \alpha (dequeue q)
        = \alpha (dequeue (Queue front kcab))
                                            -- def of q
        = \alpha (head front, Queue (tail front) kcab)
                                            -- def of dequeue (2nd equation)
        = (head front, \alpha (Queue (tail front) kcab))
                                            -- \alpha is an homomorphism

      Es bleibt zu beweisen, dass

        head front = head (\alpha q)
        \alpha (Queue (tail front) kcab) = tail (\alpha q)

        head (\alpha q)
          = head (\alpha (Queue front kcab))  -- def of q
          = head (front ++ reverse kcab)      -- def of \alpha
          = head front                        -- basic prop of head, front /= []

        \alpha (Queue (tail front) kcab)
          = tail front ++ reverse kcab        -- def of \alpha

        tail (\alpha q)
          = tail (\alpha (Queue front kcab))  -- def of q
          = tail (front ++ reverse kcab)      -- def of \alpha
          = tail front ++ reverse kcab        -- basic prop of tail, front /= []

    Fall front = [], kcab /= []:
      \alpha (dequeue q)
        = \alpha (dequeue (Queue front kcab)) -- def of q
        = \alpha (dequeue (Queue [] kcab))    -- assumption front = []
        = \alpha (dequeue (Queue (reverse kcab) []))
                                              -- def of dequeue (1st equation)
        = (head (\alpha (Queue (reverse kcab) [])),
           tail (\alpha (Queue (reverse kcab) [])))
                                              -- by reduction to already proved
                                              -- case, reverse kcab /= []
        = (head (\alpha (Queue [] kcab)), tail (\alpha (Queue [] kcab)))
                                              -- by lemma L (twice)
        = (head (\alpha q), tail (\alpha q))  -- by def of q (twice)

Beweis fuer ==:

    Zu beweisen: q == q' = \alpha q == \alpha q'

    Seien q = Queue (front kcab)
          q' = Queue (front' kcab')

    q == q'
      = Queue (front kcab) == Queue (front' kcab')
                                              -- by def of q, q'
      = toList (Queue (front kcab)) == toList (Queue (front' kcab'))
                                              -- by def of ==
      = front ++ reverse kcab == front' ++ reverse kcab'
                                              -- by def of toList

    \alpha q == \alpha q'
      = \alpha (Queue (front kcab)) == \alpha (Queue (front' kcab'))
                                              -- by def of ==
      = front ++ reverse kcab == front' ++ reverse kcab'
                                              -- by def of \alpha
-}


{- H10.1: siehe SafeMap.hs -}


{- H10.2: siehe Join.hs -}


{- H10.3: siehe Morsi.hs -}
