module Join where

import Control.Monad
import Data.List
import System.Environment (getArgs)

-- Zu implementieren
--
-- doJoin :: [String] -> [String] -> [String]
-- doJoinFormat :: [Field] -> [String] -> [String] -> [String]
-- parseFormat :: String -> Maybe [Field]

data Field = A Int | B Int

doJoin :: [String] -> [String] -> [String]
doJoin (x:xs) (y:ys)
    | ix == iy = (ix ++ rx ++ ry) : doJoin xs ys
    | ix < iy = doJoin xs (y:ys)
    | ix > iy = doJoin (x:xs) ys
  where
    (ix,rx) = break (==',') x
    (iy,ry) = break (==',') y
doJoin _ _ = []

split :: (a -> Bool) -> [a] -> [[a]]
split _ [] = []
split p xs = let (y,ys) = break p xs in
        y : case ys of [] -> []
                       (_:zs) -> split p zs


doJoinFormat :: [Field] -> [String] -> [String] -> [String]
doJoinFormat fmt = join
  where
    format fs as bs = concat $ intersperse "," $ map (\x -> case x of
        A n -> nth as n
        B n -> nth bs n) fs

    nth [] _ = ""
    nth (x:xs) 0 = x
    nth (x:xs) n = nth xs (n - 1)

    join (x:xs) (y:ys)
        | ix == iy = format fmt (split (==',') x) (split (==',') y) : join xs ys
        | ix < iy = doJoin xs (y:ys)
        | ix > iy = doJoin (x:xs) ys
      where
        ix = takeWhile (/=',') x
        iy = takeWhile (/=',') y
    join _ _ = []

parseFormat :: String -> Maybe [Field]
parseFormat = sequ . map parse . (split (==','))
  where
    parse ('A':xs) = case readInt xs of
        Nothing -> Nothing
        Just n -> Just (A n)
    parse ('B':xs) = case readInt xs of
        Nothing -> Nothing
        Just n -> Just (B n)
    parse _ = Nothing
    sequ [] = Just []
    sequ (Nothing:_) = Nothing
    sequ (Just x:xs) = case sequ xs of
        Nothing -> Nothing
        Just xs' -> Just (x:xs')

readInt :: String -> Maybe Int
readInt s =  case [x | (x,t) <- reads s, ("","") <- lex t] of
               [x] -> Just x
               _   -> Nothing

readLines :: String -> IO [String]
readLines path = liftM lines $ readFile path

main :: IO ()
main = do
    args <- getArgs
    case args of
      ["-f", fmtStr, file1, file2] | Just fmt <- parseFormat fmtStr ->
        mapM_ putStrLn =<< liftM2 (doJoinFormat fmt) (readLines file1) (readLines file2)
      [file1, file2] ->
        mapM_ putStrLn =<< liftM2 doJoin (readLines file1) (readLines file2)
      _ -> putStrLn "Invalid arguments"
