module SafeMap (SafeMap, empty, update, SafeMap.lookup) where

import Data.List (nub)

newtype SafeMap k v = SM [(k,v)]

empty :: Eq k => SafeMap k v
empty = SM []

update :: Eq k => k -> v -> SafeMap k v -> SafeMap k v
update k v (SM kvs) = SM (up k v kvs)
  where
    up k v [] = [(k,v)]
    up k v (kv'@(k',_) : kvs)
        | k == k' = (k,v) : kvs
        | otherwise = kv' : up k v kvs

lookup :: Eq k => k -> SafeMap k v -> Maybe v
lookup k (SM kvs) = Prelude.lookup k kvs


{- Abstraktionsfunktion
 -
 - alpha (SM []) = {}
 - alpha (SM ((k,v) : kvs)) = {(k,v)} \cup alpha (SM kvs)
 -}

{- Invariante: Kein Key kommt mehrfach vor -}
invar :: Eq k => SafeMap k v -> Bool
invar (SM xs) = nub ks == ks where ks = map fst xs

{- Simulationsbedingungen
 -
 - alpha_SM(empty) = {}
 - invar sm ==> alpha(update k v sm) = (alpha(sm) \ {(k',v') \in alpha(sm)| k = k'}) \cup {(k,v)}
 -
 - invar empty
 - invar sm ==> invar (update k v sm)
 -
 - Hinweis: Um eine Simulationsbedingung fuer lookup anzugeben, benoetigt
 - man eine Abstraktionsfunktion fuer den Typ Maybe v, z.B.
 -   alpha_M (Nothing) = {}
 -   alpha_M (Just x) = {x}
 -
 - Die Simulationsbedingung fuer lookup ist dann
 -
 - invar sm ==> alpha_M(lookup k sm) = {(k',v') \in alpha(sm) | k = k'}
 -}
